/* $Id: db_mysql.c,v 1.5 2005/10/30 12:31:04 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2002 The OpenTSA Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 * 
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * 
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#include "mod_tsa.h"
#include "db_mysql.h"
#include "tsa_util.h"
#include <openssl/bn.h>
#include <mysql/mysql.h>
#include <http_log.h>


/* Global module specific contexts. */
#define	MOD_TSA_MYSQL_PASSWORD	"mod_tsa_mysql_password"

/* Default options. */
#define	DEF_HOST	"localhost"
#define	DEF_PORT	3306
#define	DEF_UNIX_SOCKET	"/tmp/mysql.sock"
#define	DEF_USER	"tsa"
#define	DEF_DATABASE	"tsa"
#define	DEF_PASSPHRASE	0

/* Configuration object type. */
struct DB_mysql_conf_st
	{
	char		*host;
	unsigned int	port;
	char		*unix_socket;
	char		*user;
	char		*database;
	int		passphrase;
	};

/* Static function declarations. */
static int		db_mysql_init1(server_rec *s);
static DB		*db_mysql_init2(server_rec *s);
static int		db_mysql_connect(DB *db, server_rec *s);
static int		db_mysql_insert(DB *db, request_rec *r, TS_RESP *resp);
static int		db_mysql_disconnect(DB *db, server_rec *s);
static void		db_mysql_cleanup(DB *db, server_rec *s);
static unsigned		db_mysql_errno(DB *db);
static const char	*db_mysql_strerror(DB *db);

static void db_clear_errno(DB *db);

/* Initialising method table. */
DB_methods	db_mysql_methods =
	{
	&db_mysql_init1,
	&db_mysql_init2,
	&db_mysql_connect,
	&db_mysql_insert,
	&db_mysql_disconnect,
	&db_mysql_cleanup,
	&db_mysql_errno,
	&db_mysql_strerror
	};

/* DB connection state. */
struct DB_st
	{
	MYSQL	mysql;
	unsigned	db_errno;	/* >0 if non-mysql error occurred. */
	};

/* Error codes. */
#define	ERR_BASE	1000
#define	ERR_ALLOC_FAILURE	(ERR_BASE + 0)

/* Error strings. */
static const char *db_errors[] =
	{
	"memory allocation failed"
	};

/* Configuration handler functions. */

extern DB_mysql_conf *db_mysql_create_config(apr_pool_t *p)
	{
	DB_mysql_conf *c = (DB_mysql_conf *) 
		apr_pcalloc(p, sizeof(DB_mysql_conf));

	c->host = DEF_HOST;
	c->port = DEF_PORT;
	c->unix_socket = DEF_UNIX_SOCKET;
	c->user = DEF_USER;
	c->database = DEF_DATABASE;
	c->passphrase = DEF_PASSPHRASE;
	return c;
	}

extern const char *db_mysql_set_host(cmd_parms *cmd, void *mconfig, const char *host)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->host = apr_pstrdup(cmd->pool, host);
	return NULL;
	}

extern const char *db_mysql_set_port(cmd_parms *cmd, void *mconfig, const char *port)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->port = atoi(port);
	return NULL;
	}

extern const char *db_mysql_set_unix_socket(cmd_parms *cmd, void *mconfig,
					    const char *socket)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->unix_socket = apr_pstrdup(cmd->pool, socket);
	return NULL;
	}

extern const char *db_mysql_set_user(cmd_parms *cmd, void *mconfig, const char *user)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->user = apr_pstrdup(cmd->pool, user);
	return NULL;
	}

extern const char *db_mysql_set_database(cmd_parms *cmd, void *mconfig, 
					 const char *database)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->database = apr_pstrdup(cmd->pool, database);
	return NULL;
	}

extern const char *db_mysql_set_passphrase(cmd_parms *cmd, void *mconfig,
					   int passphrase)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->db_mysql_conf->passphrase = passphrase;
	return NULL;
	}

/* Functions for handling the database connection. */

static int db_mysql_init1(server_rec *s)
	{
	tsa_server_conf *c = (tsa_server_conf *) 
		ap_get_module_config(s->module_config, &tsa_module);

	/* Ask for the password if required. */
	if (c->db_mysql_conf->passphrase)
		{
		char password[HUGE_STRING_LEN] = "";
		char message[HUGE_STRING_LEN];
		snprintf(message, sizeof(message)-1, 
			 "Enter passphrase for database %s/%s@%s: ",
			 c->db_mysql_conf->database,
			 c->db_mysql_conf->user,
			 c->db_mysql_conf->host);
		message[sizeof(message)-1] = '\0';
		if (!tsa_get_passphrase(message, password, sizeof(password)))
			return 0;
		/* Saving password, never released. */
		apr_pool_t *p = s->process->pconf;
		apr_pool_userdata_set(apr_pstrdup(p, password), 
				      MOD_TSA_MYSQL_PASSWORD, 
				      NULL, s->process->pool);
		}

	return 1;
	}

static DB *db_mysql_init2(server_rec *s)
	{
	DB *db = (DB *) malloc(sizeof(DB));

	if (!db) return NULL;
	db_clear_errno(db);
	mysql_init(&db->mysql);

	return db;
	}

static int db_mysql_connect(DB *db, server_rec *s)
	{
	int ret;
	char *passphrase;
	apr_pool_userdata_get((void **) &passphrase, MOD_TSA_MYSQL_PASSWORD,
			      s->process->pool);
	tsa_server_conf *c = (tsa_server_conf *) 
		ap_get_module_config(s->module_config, &tsa_module);

	db_clear_errno(db);
	ret =  mysql_real_connect(&db->mysql,
				  c->db_mysql_conf->host,
				  c->db_mysql_conf->user,
				  passphrase, 
				  c->db_mysql_conf->database,
				  c->db_mysql_conf->port, 
				  c->db_mysql_conf->unix_socket, 0) != NULL;
	if (ret)
		{
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:connection to %s/%s@%s is successful",
			     tsa_module_name,
			     c->db_mysql_conf->database,
			     c->db_mysql_conf->user,
			     c->db_mysql_conf->host);
		}

	return ret;
	}

static int db_mysql_insert(DB *db, request_rec *r, TS_RESP *resp)
	{
	int ret = 0;
	PKCS7 *token = TS_RESP_get_token(resp);
	TS_TST_INFO *tst_info = TS_RESP_get_tst_info(resp);
	unsigned token_der_len;
	unsigned char *pp, *token_der = NULL;
	ASN1_INTEGER *serial;
	char *gen_time;
	BIGNUM serial_bn;
	char *serial_hex = NULL;
	char *query = NULL;
	char *esc_token_der = NULL;

	db_clear_errno(db);

	BN_init(&serial_bn);

	/* Nothing to do if there is no token in the response, return OK. */
	if (!token) return 1;

	/* Convert token to DER first. */
	token_der_len = i2d_PKCS7(token, NULL);
	pp = token_der = (unsigned char *) malloc(token_der_len);
	if (!token_der)
		{
		db->db_errno = ERR_ALLOC_FAILURE;
		goto err;
		}
	i2d_PKCS7(token, &pp);

	/* Escape the DER-encoded token. */
	esc_token_der = (char *) apr_palloc(r->pool, token_der_len*2+1);
	mysql_real_escape_string(&db->mysql, esc_token_der, 
				 token_der, token_der_len);

	/* Get the unique token id as a hex string. */
	serial = TS_TST_INFO_get_serial(tst_info);
	if (!ASN1_INTEGER_to_BN(serial, &serial_bn)
	    || !(serial_hex = BN_bn2hex(&serial_bn)))
		{
		db->db_errno = ERR_ALLOC_FAILURE;
		goto err;
		}

	/* Get the time from the token. */
	gen_time = ASN1_STRING_data(TS_TST_INFO_get_time(tst_info));

	/* Set up query string. */
	query = apr_psprintf(r->pool, "insert into token (token_id, token_date, "
			    "token_pkcs7) values('%s', '%.14s', '%s')",
			    serial_hex, gen_time, esc_token_der);

	/* Issue the query. */
	if (mysql_query(&db->mysql, query)) goto err;

	/* Log the request. */
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
		      "%s:new response (serial=0x%s) has been added to DB",
		      tsa_module_name, serial_hex);

	ret = 1;
 err:
	free(token_der);
	BN_free(&serial_bn);
	OPENSSL_free(serial_hex);
	return ret;
	}

static int db_mysql_disconnect(DB *db, server_rec *s)
	{
	db_clear_errno(db);
	mysql_close(&db->mysql);
	return 1;
	}

static void db_mysql_cleanup(DB *db, server_rec *s)
	{
	db_clear_errno(db);
	/* Nothing to do for MySQL. */
	}

extern unsigned db_mysql_errno(DB *db)
	{
	return db->db_errno != 0 ? db->db_errno : mysql_errno(&db->mysql);
	}

extern const char *db_mysql_strerror(DB *db)
	{
	return db->db_errno != 0 ? 
		db_errors[db->db_errno - ERR_BASE] : mysql_error(&db->mysql);
	}

static void db_clear_errno(DB *db)
	{
	db->db_errno = 0;
	}
