/* $Id: mod_tsa.c,v 1.15 2006/09/23 11:40:29 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2002-2005 The OpenTSA Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 * 
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * 
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#include <httpd.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_main.h>
#include <ap_config.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/ts.h>

#include "tsa_util.h"
#include "mod_tsa.h"
#include "config.h"

#define	CONTENT_TYPE_QUERY	"application/timestamp-query"
#define	CONTENT_TYPE_REPLY	"application/timestamp-reply"

/* Export module name. */
const char tsa_module_name[] = "mod_tsa";

/* Server Idendification : Modification made by JOUVE <opentsa@jouve-hdi.com>*/
char *server_id=NULL;

/* Global module specific contexts. */
#define	MOD_TSA			"mod_tsa"
#define	MOD_TSA_INITIALISED	"mod_tsa_initialised"
#define	MOD_TSA_KEY_PASSWORD	"mod_tsa_key_password"

/* ----------------------------------------------------------------------
 * Configuration option handling routines.
 * ----------------------------------------------------------------------*/

/* Default values for some of the config options. */
#define DEF_SERIAL_FILE		"conf/tsaserial"
#define DEF_CRYPTO_DEVICE	"builtin"
#define	DEF_CERTIFICATE		"conf/tsacert.pem"
#define	DEF_CERTIFICATE_CHAIN	NULL
#define	DEF_KEY			"conf/tsakey.pem"
#define	DEF_KEY_PASSPHRASE	0
#define	DEF_DEFAULT_POLICY	""
#define	DEF_POLICIES		""
#define	DEF_MESSAGE_DIGESTS	"sha1"
#define	DEF_ACCURACY_SECS	0
#define	DEF_ACCURACY_MILLISECS	0
#define	DEF_ACCURACY_MICROSECS	0
#define	DEF_PRECISION		0
#define	DEF_ORDERING		0
#define	DEF_INCLUDE_NAME	0
#define	DEF_ESS_CERT_ID_CHAIN	0
#define DEF_SERVER_ID           0 /*Modification made by JOUVE <opentsa@jouve-hdi.com>*/
#define	DEF_DB_DRIVER_NAME	"None"

/* Configuration handler functions. */

static void *create_tsa_config(apr_pool_t *p, server_rec *s)
	{
	tsa_server_conf *c =
		(tsa_server_conf *) apr_pcalloc(p, sizeof(tsa_server_conf));

	/* Setting the defaults, everything is zeroed out so far. */
	if (DEF_SERIAL_FILE)
		c->serial_file = ap_server_root_relative(p, DEF_SERIAL_FILE);
	c->crypto_device = DEF_CRYPTO_DEVICE;
	if (DEF_CERTIFICATE)
		c->certificate = ap_server_root_relative(p, DEF_CERTIFICATE);
	if (DEF_CERTIFICATE_CHAIN)
		c->certificate_chain = 
			ap_server_root_relative(p, DEF_CERTIFICATE_CHAIN);
	if (DEF_KEY)
		c->key = ap_server_root_relative(p, DEF_KEY);
	c->passphrase = DEF_KEY_PASSPHRASE;
	c->default_policy = DEF_DEFAULT_POLICY;
	c->policies = DEF_POLICIES;
	c->message_digests = DEF_MESSAGE_DIGESTS;
	c->secs = DEF_ACCURACY_SECS;
	c->millisecs = DEF_ACCURACY_MILLISECS;
	c->microsecs = DEF_ACCURACY_MICROSECS;
	c->precision = DEF_PRECISION;
	c->ordering = DEF_ORDERING;
	c->include_name = DEF_INCLUDE_NAME;
	c->ess_cert_id_chain = DEF_ESS_CERT_ID_CHAIN;
	c->server_id = DEF_SERVER_ID;/*Modification made by JOUVE <opentsa@jouve-hdi.com>*/
	c->db_driver_name = DEF_DB_DRIVER_NAME;
#ifdef TS_MYSQL
	c->db_mysql_conf = db_mysql_create_config(p);
#endif
#ifdef TS_FBSQL
	c->db_firebird_conf = db_firebird_create_config(p);
#endif
#ifdef TS_PGSQL
	c->db_pg_conf = db_pg_create_config(p);
#endif

	return c;	
}

static const char *set_serial_file(cmd_parms *cmd, void *mconfig, const char *serial)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->serial_file = ap_server_root_relative(cmd->pool, serial);
	return NULL;
	}

static const char *set_crypto_device(cmd_parms *cmd, void *mconfig,
				     const char *crypto_device)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->crypto_device = apr_pstrdup(cmd->pool, crypto_device);
	return NULL;
	}

static const char *set_certificate(cmd_parms *cmd, void *mconfig, const char *cert)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->certificate = ap_server_root_relative(cmd->pool, cert);
	return NULL;
	}

static const char *set_certificate_chain(cmd_parms *cmd, void *mconfig, 
					 const char *chain)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->certificate_chain = ap_server_root_relative(cmd->pool, chain);
	return NULL;
	}

static const char *set_tsa_key(cmd_parms *cmd, void *mconfig, const char *key)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->key = ap_server_root_relative(cmd->pool, key);
	return NULL;
	}

static const char *set_tsa_key_passphrase(cmd_parms *cmd, void *mconfig, 
					  int passphrase)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->passphrase = passphrase;
	return NULL;
	}

static const char *set_default_policy(cmd_parms *cmd, void *mconfig, 
				      const char *policy)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->default_policy = apr_pstrdup(cmd->pool, policy);
	return NULL;
	}

static const char *set_policies(cmd_parms *cmd, void *mconfig, const char *policies)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->policies = apr_pstrdup(cmd->pool, policies);
	return NULL;
	}

static const char *set_message_digests(cmd_parms *cmd, void *mconfig, const char *mds)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->message_digests = apr_pstrdup(cmd->pool, mds);
	return NULL;
	}

static const char *set_accuracy(cmd_parms *cmd, void *mconfig, 
				const char *secs, const char *millisecs, const char *microsecs)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->secs = atoi(secs);
	cfg->millisecs = atoi(millisecs);
	cfg->microsecs = atoi(microsecs);
	return NULL;
	}

static const char *set_clock_precision_digits(cmd_parms *cmd, void *mconfig, 
					      const char *precision)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->precision = atoi(precision);
	return NULL;
	}

static const char *set_ordering(cmd_parms *cmd, void *mconfig, int ordering)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->ordering = ordering;
	return NULL;
	}

static const char *set_include_name(cmd_parms *cmd, void *mconfig, int inc_name)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->include_name = inc_name;
	return NULL;
	}

static const char *set_ess_cert_id_chain(cmd_parms *cmd, void *mconfig, int ids)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);

	cfg->ess_cert_id_chain = ids;
	return NULL;
	}
/*Modification made by JOUVE <opentsa@jouve-hdi.com>*/
static const char * set_server_id(cmd_parms *cmd, void *mconfig, const char *server_id)
{
  tsa_server_conf *cfg = (tsa_server_conf *)
    ap_get_module_config(cmd->server->module_config, &tsa_module);
  
  cfg->server_id = apr_pstrdup(cmd->pool, server_id);
  return NULL;

}

static const char *set_db_module(cmd_parms *cmd, void *mconfig, const char *driver)
	{
	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(cmd->server->module_config, &tsa_module);
	
	cfg->db_driver_name = apr_pstrdup(cmd->pool, driver);
	return NULL;
	}

/* ----------------------------------------------------------------------
 * Module initialisation.
 * ----------------------------------------------------------------------*/

static int conf_set_serial(server_rec *s, apr_pool_t *p, 
			   TS_RESP_CTX *ctx, char *serial)
	{
	int ret = 0;
	apr_file_t *serial_fp = NULL;
	apr_status_t rc;

	rc = apr_file_open(&serial_fp, serial, 
		APR_READ | APR_WRITE, APR_OS_DEFAULT, p);
 
	if (rc != APR_SUCCESS)
		{
		/* Try to create and initialise it with 0. */
		rc = apr_file_open(&serial_fp, serial, 
			APR_READ | APR_WRITE | APR_CREATE, APR_OS_DEFAULT, p);
		if (rc != APR_SUCCESS) goto end;
		/* Even number of digits must be written. */
		apr_file_puts("00\n", serial_fp);
		apr_file_flush(serial_fp);
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:new serial file is created with serial "
			     " number 0: %s", tsa_module_name, serial);
		}
	else
		{
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:serial file is re-used: %s",
			     tsa_module_name, serial);
		}
	TS_RESP_CTX_set_serial_cb(ctx, tsa_serial_cb, serial_fp);

#ifdef TS_DEBUG
	if (tsa_error()) goto end;
#endif

	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not create serial file: %s",
			     tsa_module_name, serial);
	return ret;
	}

static int conf_set_crypto_device(server_rec *s, char *crypto_device)
	{
	int ret = 0;

	if (!TS_CONF_set_default_engine(crypto_device)) goto end;

	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:crypto device is set to: %s",
		     tsa_module_name, crypto_device);
	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not set crypto device: %s",
			     tsa_module_name, crypto_device);
	return ret;
	}

static int conf_set_signer_cert(server_rec *s, TS_RESP_CTX *ctx,
				const char *cert)
	{
	int ret = 0;
	X509 *cert_obj = NULL;
		
	if (!(cert_obj = tsa_load_cert(cert, FORMAT_PEM))) goto end;
	if (!TS_RESP_CTX_set_signer_cert(ctx, cert_obj)) goto end;
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:X.509 signer certificate is loaded: %s",
		     tsa_module_name, cert);
	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not load X.509 certificate: %s",
			     tsa_module_name, cert);
	X509_free(cert_obj);
	return ret;
	}

static int conf_set_certs(server_rec *s, TS_RESP_CTX *ctx, char *chain)
	{
	int ret = 0;
	STACK_OF(X509) *certs = NULL;
		
	/* Certificate chain is optional. */
	if (!chain) return 1;

	if (!(certs = tsa_load_certs(chain))) goto end;
	if (!TS_RESP_CTX_set_certs(ctx, certs)) goto end;
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:X.509 signer certificate chain is loaded: %s",
		     tsa_module_name, chain);
	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not load X.509 certificate chain: %s",
			     tsa_module_name, chain);
	sk_X509_pop_free(certs, X509_free);
	return ret;
	}

static int conf_set_signer_key(server_rec *s, TS_RESP_CTX *ctx, char *key,
			       char *passwd)
	{
	int ret = 0;
	EVP_PKEY *key_obj = NULL;

	if (!(key_obj = tsa_load_key(key, FORMAT_PEM, passwd))) goto end;
	if (!TS_RESP_CTX_set_signer_key(ctx, key_obj)) goto end;
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:signer key is initialised: %s",
		     tsa_module_name, key);

	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not load private key: %s",
			     tsa_module_name, key);
	EVP_PKEY_free(key_obj);
	return ret;
	}

static int conf_set_def_policy(server_rec *s, TS_RESP_CTX *ctx,
			       char *def_policy)
	{
	int ret = 0;
	ASN1_OBJECT *oid = NULL;
	char buffer[1024];

	if (!(oid = OBJ_txt2obj(def_policy, 0))) goto end;
	if (!TS_RESP_CTX_set_def_policy(ctx, oid)) goto end;
	if (!OBJ_obj2txt(buffer, sizeof(buffer), oid, 0)) goto end;
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:default policy is set to: %s",
		     tsa_module_name, buffer);

	ret = 1;
 end:
	if (!ret)
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:invalid default policy: %s",
			     tsa_module_name, def_policy);
	ASN1_OBJECT_free(oid);
	return ret;
	}

static int conf_set_policies(server_rec *s, apr_pool_t *p,
			     TS_RESP_CTX *ctx, const char *policies)
	{
	int ret = 0;
	char *policy;
	char buf[1024];

	while (*(policy = ap_getword_conf(p, &policies)) != '\0')
		{
		ASN1_OBJECT *objtmp;
		if (!(objtmp = OBJ_txt2obj(policy, 0)) 
		    || !TS_RESP_CTX_add_policy(ctx, objtmp)
		    || !OBJ_obj2txt(buf, sizeof(buf), objtmp, 0))
			{
			ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
				     "%s:could not set policy: %s",
				     tsa_module_name, policy);
			ASN1_OBJECT_free(objtmp);
			goto end;
			}
		ASN1_OBJECT_free(objtmp);
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:acceptable policy set includes: %s",
			     tsa_module_name, buf);
		}
	
	ret = 1;
 end:
	return ret;
	}

static int conf_set_digests(server_rec *s, apr_pool_t *p, TS_RESP_CTX *ctx,
			    const char *mds)
	{
	int ret = 0;
	char *md;

	while (*(md = ap_getword_conf(p, &mds)) != '\0')
		{
		const EVP_MD *md_obj;
		if (!(md_obj = EVP_get_digestbyname(md))
		    || !TS_RESP_CTX_add_md(ctx, md_obj))
			{
			ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
				     "%s:could not set message digest: %s",
				     tsa_module_name, md);
			goto end;
			}
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:acceptable message digest set includes: %s",
			     tsa_module_name, md);
		}
	ret = 1;
 end:
	return ret;
	}

static int conf_set_accuracy(server_rec *s, TS_RESP_CTX *ctx, int secs,
			     int millisecs, int microsecs)
	{
	if (!TS_RESP_CTX_set_accuracy(ctx, secs, millisecs, microsecs))
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not set accuracy to: %d, %d, %d",
			     tsa_module_name, secs, millisecs, microsecs);
		return 0;
		}
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:accuracy is set to: %d secs, %d millisecs, "
		     "%d microsecs", 
		     tsa_module_name, secs, millisecs, microsecs);

	return 1;
	}

static int conf_set_clock_precision_digits(server_rec *s, TS_RESP_CTX *ctx,
					   int precision)
	{
	if (!TS_RESP_CTX_set_clock_precision_digits(ctx, precision))
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:could not set clock precision to: %d",
			     tsa_module_name, precision);
		return 0;
		}
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:clock precision is set to: %d", 
		     tsa_module_name, precision);

	return 1;
	}

static int conf_set_ordering(server_rec *s, TS_RESP_CTX *ctx, int ordering)
	{
	if (ordering)
		TS_RESP_CTX_add_flags(ctx, TS_ORDERING);
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:ordering is switched %s", 
		     tsa_module_name, ordering ? "On" : "Off");
	return 1;
	}

static int conf_set_tsa_name(server_rec *s, TS_RESP_CTX *ctx, int inc_name)
	{
	if (inc_name)
		TS_RESP_CTX_add_flags(ctx, TS_TSA_NAME);
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:TSA name inclusion is switched %s", 
		     tsa_module_name, inc_name ? "On" : "Off");
	return 1;
	}

static int conf_set_ess_cert_id_chain(server_rec *s, TS_RESP_CTX *ctx, 
				      int ess_ids)
	{
	if (ess_ids)
		TS_RESP_CTX_add_flags(ctx, TS_ESS_CERT_ID_CHAIN);
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:ESS certificate id chain inclusion is switched %s", 
		     tsa_module_name, ess_ids ? "On" : "Off");
	return 1;
	}

/* Configuration of Server Identification : Modification made by JOUVE <opentsa@jouve-hdi.com> */
static int conf_set_server_id(server_rec *s, char *cfg_server_id)
{
  if (cfg_server_id==NULL) 
	return 1;

  server_id = cfg_server_id;
  ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
	       "%s:server identification is set to: %s",
	       tsa_module_name, server_id);
  return 1;
}

static int conf_set_db_driver(server_rec *s, tsa_server_conf *cfg)
	{
	int ret = 0;

	if (strcmp(cfg->db_driver_name, "None") == 0)
		{
		cfg->db_driver = NULL;
		}
#ifdef TS_MYSQL
	else if (strcmp(cfg->db_driver_name, "MySQL") == 0)
		{
		cfg->db_driver = &db_mysql_methods;
		}
#endif
#ifdef TS_FBSQL
        else if (strcmp(cfg->db_driver_name, "FireBird") == 0)
                {
                cfg->db_driver = &db_firebird_methods;
                }
#endif
#ifdef TS_PGSQL
	else if (strcmp(cfg->db_driver_name, "PostgreSQL") == 0)
		{
		cfg->db_driver = &db_pg_methods;
		}
#endif
	else
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:unknown database driver: %s",
			     tsa_module_name, cfg->db_driver_name);
		goto end;
		}
	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:database driver is set to: %s", 
		     tsa_module_name, cfg->db_driver_name);
	ret = 1;
end:
	return ret;
	}

static int conf_db_init(server_rec *s, tsa_server_conf *cfg)
	{
	/* Initialise database backend, second round. */
	if (!(cfg->db = (*cfg->db_driver->db_init2)(s)))
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:cannot initialise database",
			     tsa_module_name);
		return 0;
		}
	/* Connect to the database. */
	if (!(*cfg->db_driver->db_connect)(cfg->db, s))
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:cannot connect to database:%d:%s",
			     tsa_module_name,
			     (*cfg->db_driver->db_errno)(cfg->db),
			     (*cfg->db_driver->db_strerror)(cfg->db));
		return 0;
		}

	return 1;
	}

/* ----------------------------------------------------------------------
 *  Module initialiser method:
 *	- Initialises OpenSSL.
 *	- Creates and initialises the shared TS_RESP_CTX object.
 * ----------------------------------------------------------------------*/

static void tsa_module_init1(server_rec *s, apr_pool_t *p)
	{
#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init1: start");
#endif

	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(s->module_config, &tsa_module);

	/* config ? */
	if (!cfg) goto err;

	/* Shall we ask for private key passphrase? */
	if (cfg->passphrase)
		{
		char password[HUGE_STRING_LEN] = "";
		char *message;

		message = apr_psprintf(p, "Enter passphrase for %s: ", cfg->key);
		if (!tsa_get_passphrase(message, password, sizeof(password)))
			{
#ifdef TS_DEBUG
			ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
				"tsa_module_init1: %s", password);
#endif
			goto err;
			}

		/* Saving password, never released. */
#ifdef TS_DEBUG
		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
			"tsa_module_init1: %s", password);
#endif
		apr_pool_userdata_set(apr_pstrdup(p, password), MOD_TSA_KEY_PASSWORD, NULL, s->process->pool);
		}

	/* Setting the database backend driver. */
	if (!conf_set_db_driver(s, cfg)) goto err;

	/* Initialise database backend if used, first round. */
	if (cfg->db_driver && !(*cfg->db_driver->db_init1)(s))
		{
		ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
			     "%s:cannot initialise database",
			     tsa_module_name);
		goto err;
		}

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init1: end");
#endif

	return;
 err:

#ifdef TS_DEBUG
	if (tsa_error())
		ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s, 
			tsa_error());
#endif

	ap_log_error(APLOG_MARK, APLOG_EMERG | APLOG_NOERRNO, 0, s, 
		     "exiting, fatal error during mod_tsa initialisation.");
	exit(1);
	}

static void tsa_module_init2(server_rec *s, apr_pool_t *p)
	{
#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init2: start");
#endif

	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(s->module_config, &tsa_module);

	char *passwd = NULL;

        apr_pool_userdata_get((void **)&passwd, MOD_TSA_KEY_PASSWORD,
			      s->process->pool);

	/* Initialising OpenSSL. */
	tsa_openssl_init();

        /* Initialising Server Identification : Modification made by JOUVE <opentsa@jouve-hdi.com> */
        if (!conf_set_server_id(s, cfg->server_id))
	  goto err;

	/* Initialise pid, needed for OpenSSL cleanup. */
	cfg->init_pid = (unsigned long) getpid();

	/* Initialise password. */
	cfg->key_passwd = apr_pstrdup(p, passwd);

	/* Initialise the TSA data. */
	/* Memory allocated here will be freed by tsa_child_exit. */
	if (!(cfg->resp_ctx = TS_RESP_CTX_new())) goto err;

	/* Adding the response context object to the context of mod_tsa. */
	apr_pool_userdata_set(cfg->resp_ctx, MOD_TSA, NULL, s->process->pool);

	/* Setting serial number provider callback. */
	if (!conf_set_serial(s, p, cfg->resp_ctx, cfg->serial_file))
		goto err;

	/* Setting crypto device for signing operations. */
	if (!conf_set_crypto_device(s, cfg->crypto_device))
		goto err;

	/* Setting TSA signer certificate. */
	if (!conf_set_signer_cert(s, cfg->resp_ctx, cfg->certificate))
		goto err;

	/* Setting TSA signer certificate. */
	if (!conf_set_certs(s, cfg->resp_ctx, cfg->certificate_chain))
		goto err;

	/* Setting TSA signer private key. */
	if (!conf_set_signer_key(s, cfg->resp_ctx, cfg->key, cfg->key_passwd))
		goto err;

	/* Setting default policy OID. */
	if (!conf_set_def_policy(s, cfg->resp_ctx, cfg->default_policy))
		goto err;

	/* Setting acceptable policy OIDs. */
	if (!conf_set_policies(s, p, cfg->resp_ctx, cfg->policies))
		goto err;

	/* Setting the acceptable one-way hash algorithms. */
	if (!conf_set_digests(s, p, cfg->resp_ctx, cfg->message_digests))
		goto err;

	/* Setting guaranteed time stamp accuracy. */
	if (!conf_set_accuracy(s, cfg->resp_ctx, cfg->secs,
			       cfg->millisecs, cfg->microsecs))
		goto err;

	/* Setting the time precision included in the time stamp token. */
	if (!conf_set_clock_precision_digits(s, cfg->resp_ctx, cfg->precision))
		goto err;

	/* Setting the ordering flag if requested. */
	if (!conf_set_ordering(s, cfg->resp_ctx, cfg->ordering))
		goto err;

	/* Setting the TSA name required flag if requested. */
	if (!conf_set_tsa_name(s, cfg->resp_ctx, cfg->include_name))
		goto err;

	/* Setting the ESS cert id chain flag if requested. */
	if (!conf_set_ess_cert_id_chain(s, cfg->resp_ctx,
					cfg->ess_cert_id_chain)) 
		goto err;

	/* Setting the database backend driver. */
	if (!conf_set_db_driver(s, cfg)) goto err;

	/* Initialise database backend if used. */
	if (cfg->db_driver && !conf_db_init(s, cfg)) goto err;

	ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
		     "%s:module has started successfully", tsa_module_name);

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init2: end");
#endif

	return;
 err:
	tsa_print_errors(s);
	/* There is not much left to do... */
	ap_log_error(APLOG_MARK, APLOG_EMERG | APLOG_NOERRNO, 0, s,
		     "exiting, fatal error during mod_tsa initialisation.");
	/* No memory clean up, because we are bailing out... */
	exit(1);
	}

static apr_status_t tsa_cleanup(void *s_);

static void tsa_module_init(server_rec *s, apr_pool_t *p)
	{
#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init: start");
#endif

	/* Has this method been called yet? */
	void *init = NULL;
	apr_pool_userdata_get(&init, MOD_TSA_INITIALISED, s->process->pool);

	if (init)
		{
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:re-initialization started", tsa_module_name);
		tsa_module_init2(s, p);
		}
	else
		{
		ap_log_error(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, 0, s,
			     "%s:mod_tsa initialization started, "
			     "MySQL support:"
#ifdef TS_MYSQL
			     "enabled"
#else
			     "disabled"
#endif
			     ", FireBird support:"
#ifdef TS_FBSQL
			     "enabled"
#else
			     "disabled"
#endif
			     ", PostgreSQL support:"
#ifdef TS_PGSQL
			     "enabled"
#else
			     "disabled"
#endif
			     ,tsa_module_name);
		tsa_module_init1(s, p);
		apr_pool_cleanup_register(p, s, tsa_cleanup, apr_pool_cleanup_null);
		apr_pool_userdata_set((const void*) 1, MOD_TSA_INITIALISED, NULL, s->process->pool);
		}

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_module_init: end");
#endif

	}

static int tsa_post_config(apr_pool_t *pconf,apr_pool_t *plog,
                                 apr_pool_t *ptemp,server_rec *s)
	{
		tsa_module_init(s, pconf);
		return OK;
	}


/* ----------------------------------------------------------------------
 * Child exit routine.
 * ----------------------------------------------------------------------*/

static void tsa_child_exit(server_rec *s, apr_pool_t *p)
	{
#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_child_exit: start");
#endif

	tsa_server_conf *cfg = (tsa_server_conf *) 
		ap_get_module_config(s->module_config, &tsa_module);

	/* Clean up response context. */
	TS_RESP_CTX_free(cfg->resp_ctx);
	
	/* Disconnect and cleanup database. */
	if (cfg->db_driver && cfg->db)
		{
		if (!(*cfg->db_driver->db_disconnect)(cfg->db, s))
			{
			ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, s,
				     "%s:cannot disconnect from database:%d:%s",
				     tsa_module_name, 
				     (*cfg->db_driver->db_errno)(cfg->db),
				     (*cfg->db_driver->db_strerror)(cfg->db));
			}
		(*cfg->db_driver->db_cleanup)(cfg->db, s);
		}

	/* Clean up OpenSSL. */
	tsa_openssl_cleanup(s, cfg->init_pid);

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_child_exit: end");
#endif

	}

static apr_status_t tsa_cleanup(void *s_)
	{
	server_rec *s = (server_rec *) s_;

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_cleanup: start");
#endif

	tsa_child_exit(s, s->process->pconf);

#ifdef TS_DEBUG
	ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, s,
		     "tsa_cleanup: end");
#endif

	return APR_SUCCESS;
	}

/* ----------------------------------------------------------------------
 * Request handling routines.
 * ----------------------------------------------------------------------*/

AP_DECLARE(apr_status_t) ap_send_http_header(request_rec * const r)
{
    /* (similar to ap_rflush() but returns (apr_status_t) instead of (int)) */
    apr_bucket_alloc_t * const bucket_alloc = r->connection->bucket_alloc;
    apr_bucket_brigade * const bb = apr_brigade_create(r->pool, bucket_alloc);
    APR_BRIGADE_INSERT_HEAD(bb, apr_bucket_flush_create(bucket_alloc));
    return ap_pass_brigade(r->output_filters, bb);
}


static int tsa_handler(request_rec *r)
	{
#ifdef TS_DEBUG
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			"tsa_handler: start");
#endif

	int ret;
	tsa_server_conf *c = (tsa_server_conf *) 
		ap_get_module_config(r->server->module_config, &tsa_module);
	char	*req;
	int	req_len, len, content_length;
	BIO	*req_bio = NULL;
	unsigned char	*pp, *response;
	int	response_len;
	TS_RESP *response_obj = NULL;
	const char	*ct;

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
		      "%s:request handler is called", tsa_module_name);

	/* Checking for the only allowed POST method. */
	if (r->method_number != M_POST)
		{
		r->allowed = 1 << M_POST;
		ret = DECLINED;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:not POST method was used in request",
			      tsa_module_name);
		goto end;
		}

	/* Checking content type. */
	if (!(ct = apr_table_get(r->headers_in, "content-type"))
	    || strcmp(ct, CONTENT_TYPE_QUERY) != 0)
		{
		ret = HTTP_BAD_REQUEST;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:bad Content-Type in request", 
			      tsa_module_name);
		goto end;
		}

	/* Reading the request body. */
	if ((ret = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)) != OK)
		{
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:bad request: no Content-Length?",
			      tsa_module_name);
		goto end;
		}
	if (!ap_should_client_block(r))
		{
		ret = HTTP_BAD_REQUEST;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:bad Content-Type in request", 
			      tsa_module_name);
		goto end;
		}

	/* We must have a Content-Length, allocate memory. */
	content_length = r->remaining;
	req = (char *) apr_pcalloc(r->pool, content_length);
	ap_soft_timeout("mod_tsa: reading request", r);
	for (req_len = 0;
	     (len = ap_get_client_block(r, req + req_len,
					content_length - req_len)) > 0;
	     req_len += len) ;
	ap_kill_timeout(r);
	if (content_length != req_len)
		{
		ret = HTTP_REQUEST_TIME_OUT;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:request timed out", tsa_module_name);
		goto end;
		}

	/* Creating response. */
	if (!(req_bio = BIO_new_mem_buf(req, req_len))
	    || !(response_obj = TS_RESP_create_response(c->resp_ctx, req_bio)))
		{
		ret = HTTP_INTERNAL_SERVER_ERROR;
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:could not create response", tsa_module_name);
		goto end;
		}
	
	/* Add the time stamp token to the database first 
	   if database is used. */
	if (c->db_driver && !(*c->db_driver->db_insert)(c->db, r, response_obj))
		{
		ret = HTTP_INTERNAL_SERVER_ERROR;
		ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r,
			      "%s:could not add token to database:%d:%s",
			      tsa_module_name, 
			      (*c->db_driver->db_errno)(c->db), 
			      (*c->db_driver->db_strerror)(c->db));
		goto end;
		}

	/* Sending response. */
	response_len = i2d_TS_RESP(response_obj, NULL);
	pp = response = (unsigned char *) apr_pcalloc(r->pool, response_len);
	i2d_TS_RESP(response_obj, &pp);
	r->content_type = CONTENT_TYPE_REPLY;
	ap_set_content_length(r, response_len);

    	ap_send_http_header(r);
	/* We know that we are not handling a HEAD request, so
	   go on with the body. */
	ap_soft_timeout("mod_tsa: sending response", r);
	len = ap_rwrite(response, response_len, r);
	ap_kill_timeout(r);
	if (len != response_len)
		{
		ret = HTTP_GONE;	/* Is this the appropriate error? */
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
			      "%s:could not send response in full", 
			      tsa_module_name);
		goto end;
		}

	tsa_print_errors(r->server);

	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
		      "%s:request handler finished successfully", 
		      tsa_module_name);
	ret = OK;
 end:
	BIO_free_all(req_bio);
	TS_RESP_free(response_obj);

#ifdef TS_DEBUG
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r,
		      "tsa_handler: end");
#endif

	return ret;
	}

/* ----------------------------------------------------------------------
 * Module information structures.
 * ----------------------------------------------------------------------*/

/* TSA config file commands. */
static const command_rec tsa_cmds[] =
{
    AP_INIT_TAKE1("TSASerialFile", set_serial_file, NULL, RSRC_CONF, 
	"the file which contains the next serial number"),
    AP_INIT_TAKE1("TSACryptoDevice", set_crypto_device, NULL, RSRC_CONF, 
	"engine id to use for cryptographic operations"),
    AP_INIT_TAKE1("TSACertificate", set_certificate, NULL, RSRC_CONF, 
	"the certificate of the TSA"),
    AP_INIT_TAKE1("TSACertificateChain", set_certificate_chain, NULL, RSRC_CONF,
	"certificate chain for the TSA excluding the TSA certificate"),
    AP_INIT_TAKE1("TSAKey", set_tsa_key, NULL, RSRC_CONF, 
	"private key of the TSA"),
    AP_INIT_FLAG("TSAKeyPassPhrase", set_tsa_key_passphrase, NULL, RSRC_CONF,
     "specifies if the private key is protected by a passphrase"),
    AP_INIT_TAKE1("TSADefaultPolicy", set_default_policy, NULL, RSRC_CONF,
     "the OID of the default TSA service policy"),
    AP_INIT_RAW_ARGS("TSAPolicies", set_policies, NULL, RSRC_CONF,
     "list of service policy OIDs acceptable by the TSA excluding the default"),
    AP_INIT_RAW_ARGS("TSAMessageDigests", set_message_digests, NULL, RSRC_CONF,
     "list of acceptable message digest algorithms"),
    AP_INIT_TAKE3("TSAAccuracy", set_accuracy, NULL, RSRC_CONF,
     "the accuracy of the time source in seconds, milliseconds "
     "and microseconds"),
    AP_INIT_TAKE1("TSAClockPrecisionDigits", set_clock_precision_digits, NULL, RSRC_CONF,
     "the number of digits included in the time field after the dot."),
    AP_INIT_FLAG("TSAOrdering", set_ordering, NULL, RSRC_CONF,
     "specifies if the TSA provides ordering semantics of time stamps"),
    AP_INIT_FLAG("TSAIncludeName", set_include_name, NULL, RSRC_CONF,
     "specified if the TSA is to include its name in time stamp tokens"),
    AP_INIT_FLAG("TSAESSCertIdChain", set_ess_cert_id_chain, NULL, RSRC_CONF,
     "the TSA adds ESS certificate ids for each certificate of the chain, "
     "not just for the signing certificate"),
    AP_INIT_TAKE1("TSAServerId",set_server_id,NULL,RSRC_CONF,/*Modification made by JOUVE <opentsa@jouve-hdi.com>*/
		  "the server id to concatenate with the serial number"),
    AP_INIT_TAKE1("TSADBModule", set_db_module, NULL, RSRC_CONF,
     "the database backend to use for logging the requests: None"
#ifdef TS_MYSQL
     ", MySQL"
#endif
#ifdef TS_FBSQL
     ", FireBird"
#endif
#ifdef TS_PGSQL
     ", PostgreSQL"
#endif
    ),
#ifdef TS_MYSQL
    AP_INIT_TAKE1("TSAMySQLHost", db_mysql_set_host, NULL, RSRC_CONF, 
	"DNS name or IP address of the MySQL host"),
    AP_INIT_TAKE1("TSAMySQLPort", db_mysql_set_port, NULL, RSRC_CONF, 
	"port of the MySQL server to connect to"),
    AP_INIT_TAKE1("TSAMySQLUnixSocket", db_mysql_set_unix_socket, NULL, RSRC_CONF, 
	"unix domain socket of the MySQL server to connect to"),
    AP_INIT_TAKE1("TSAMySQLUser", db_mysql_set_user, NULL, RSRC_CONF, 
	"MySQL user name to connect as"),
    AP_INIT_TAKE1("TSAMySQLDatabase", db_mysql_set_database, NULL, RSRC_CONF, 
	"name of the MySQL database to use for logging"),
    AP_INIT_FLAG("TSAMySQLPassPhrase", db_mysql_set_passphrase, NULL, RSRC_CONF,
     "specifies if a passphrase is needed for the database connection"),
#endif
#ifdef TS_FBSQL
    AP_INIT_TAKE1("TSAFireBirdHost", db_firebird_set_host, NULL, RSRC_CONF, 
	"DNS name or IP address of the FireBird host"),
    AP_INIT_TAKE1("TSAFireBirdPort", db_firebird_set_port, NULL, RSRC_CONF, 
	"port of the FireBird server to connect to"),
    AP_INIT_TAKE1("TSAFireBirdUnixSocket", db_firebird_set_unix_socket, NULL, RSRC_CONF, 
	"unix domain socket of the FireBird server to connect to"),
    AP_INIT_TAKE1("TSAFireBirdUser", db_firebird_set_user, NULL, RSRC_CONF, 
	"FireBird user name to connect as"),
    AP_INIT_TAKE1("TSAFireBirdDatabase", db_firebird_set_database, NULL, RSRC_CONF, 
	"name of the FireBird database to use for logging"),
    AP_INIT_FLAG("TSAFireBirdPassPhrase", db_firebird_set_passphrase, NULL, RSRC_CONF,
     "specifies if a passphrase is needed for the database connection"),
#endif
#ifdef TS_PGSQL
    AP_INIT_TAKE1("TSAPostgreSQLHost", db_pg_set_host, NULL, RSRC_CONF,
	"DNS name or IP address of the PostgreSQL host"),
    AP_INIT_TAKE1("TSAPostgreSQLPort", db_pg_set_port, NULL, RSRC_CONF,
	"port of the PostgreSQL server to connect to"),
    AP_INIT_TAKE1("TSAPostgreSQLUser", db_pg_set_user, NULL, RSRC_CONF,
	"PostgreSQL user name to connect as"),
    AP_INIT_TAKE1("TSAPostgreSQLDatabase", db_pg_set_database, NULL, RSRC_CONF,
	"name of the PostgreSQL database to use for logging"),
    AP_INIT_FLAG("TSAPostgreSQLPassPhrase", db_pg_set_passphrase, NULL, RSRC_CONF,
     "specifies if a passphrase is needed for the database connection"),
#endif
    {NULL}
};

static void tsa_register_hooks(apr_pool_t *p)
{

#ifdef TS_DEBUG
    ap_log_perror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, p,
			     "tsa_register_hooks: start");
#endif

    ap_hook_post_config(tsa_post_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(tsa_handler, NULL, NULL, APR_HOOK_MIDDLE);

#ifdef TS_DEBUG
    ap_log_perror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, p,
			     "tsa_register_hooks: end");
#endif

}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA tsa_module = {
    STANDARD20_MODULE_STUFF, 
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    create_tsa_config,     /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    tsa_cmds,              /* table of config file commands       */
    tsa_register_hooks     /* register hooks                      */
};
