/* $Id: mod_tsa.h,v 1.8 2006/09/23 10:56:47 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2002 The OpenTSA Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 * 
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * 
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#ifndef	__MOD_TSA_H__
#define	__MOD_TSA_H__

#include <httpd.h>
#include <apr_strings.h>

AP_DECLARE(apr_status_t) ap_send_http_header(request_rec *r);

#define ap_soft_timeout(a,b)
#define ap_kill_timeout(a)

#include "config.h"

#include "db.h"
#ifdef TS_MYSQL
#include "db_mysql.h"
#endif
#ifdef TS_FBSQL
#include "db_firebird.h"
#endif
#ifdef TS_PGSQL
#include "db_postgresql.h"
#endif
#include <openssl/ts.h>

/* Module table. */
extern module AP_MODULE_DECLARE_DATA tsa_module;

/* Global module name. */
extern const char tsa_module_name[];

/*Server Identification : Modification made by JOUVE <opentsa@jouve-hdi.com */
extern char *server_id;

/* Module config structure. */
typedef struct 
	{
	char	*serial_file;		/* Contains the unique serial ID. */
	char	*crypto_device;		/* Engine ID for signing. */
	char	*certificate;		/* TSA signing certificate file. */
	char	*certificate_chain;	/* The CA certs for 'certificate'. */
	char	*key;			/* TSA private key file. */
	int	passphrase;		/* Is there a passphrase for key? */
	char	*default_policy;	/* Default TSA policy OID. */
	char	*policies;		/* Acceptable policy OIDs. */
	char	*message_digests;	/* Acceptable message digests. */
	int	secs, millisecs, microsecs;	/* TSA accuracy. */
	int	precision;		/* fraction of seconds precision. */
	int	ordering;		/* Is ordering defined for tokens? */
	int	include_name;	/* Should TSA name be included in response? */
	int	ess_cert_id_chain;	/* Shall we include cert IDs? */
	char    *server_id;             /*Server Identification : Modification made by JOUVE <opentsa@jouve-hdi.com*/            
	char	*db_driver_name;	/* Name of the database backend. */
#ifdef TS_MYSQL
	DB_mysql_conf	*db_mysql_conf;	/* MySQL database config options. */
#endif
#ifdef TS_FBSQL
	DB_firebird_conf *db_firebird_conf; /* FireBird database config. */
#endif
#ifdef TS_PGSQL
	DB_pg_conf *db_pg_conf;		/* PostgreSQL database config. */
#endif

	/* The fields below are initialised by tsa_module_init. */
	TS_RESP_CTX	*resp_ctx;
	/* The pid of the process that initialises OpenSSL. */
	unsigned long	init_pid;
	/* Private key pass phrase. */
	char		*key_passwd;
	/* Database methods and handle for storing tokens. */
	DB_methods	*db_driver;
	DB		*db;
	} tsa_server_conf;

#endif	/* __MOD_TSA_H__ */
