/* $Id: tsa_util.h,v 1.5 2005/10/29 13:57:41 zglozik Exp $ */
/* ====================================================================
 * Copyright (c) 2002 The OpenTSA Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * 4. The name "OpenTSA" must not be used to endorse or promote products
 *    derived from this software without prior written permission.
 *    For written permission, please contact zglozik@stones.com.
 * 
 * 5. Products derived from this software may not be called "OpenTSA"
 *    nor may "OpenTSA" appear in their names without prior written
 *    permission of the OpenTSA Project.
 * 
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenTSA Project.
 *    (http://www.opentsa.org/)"
 * 
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenTSA PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * 
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit (http://www.OpenSSL.org/).
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).
 * This product includes software written by Tim Hudson (tjh@cryptsoft.com).
 */

#ifndef	__TSA_UTIL_H__
#define	__TSA_UTIL_H__

#include <httpd.h>
#include <apr_portable.h>
#include <openssl/x509.h>
#include <openssl/ts.h>

/* OpenSSL-related utility functions. */

extern void tsa_openssl_init(void);

extern void tsa_openssl_cleanup(server_rec *s, unsigned long init_pid);

extern void tsa_print_errors(server_rec *s);

#define FORMAT_ASN1     1
#define FORMAT_PEM      2
#define FORMAT_PKCS12   3

/* format can be one of: FORMAT_ASN1, FORMAT_PEM, FORMAT_PKCS12 */
extern X509 *tsa_load_cert(const char *file, int format);

extern STACK_OF(X509) *tsa_load_certs(const char *file);

/* format can be one of: FORMAT_ASN1, FORMAT_PEM, FORMAT_PKCS12 */
extern EVP_PKEY *tsa_load_key(const char *file, int format, const char *pass);

extern ASN1_INTEGER *tsa_serial_cb(TS_RESP_CTX *ctx, void *data);

/* Change stderr for the console and ask for a passphrase. */
extern int tsa_get_passphrase(const char *message, char *buffer, unsigned size);

#ifdef TS_DEBUG
extern const char* tsa_error(void);
#define TSA_ERROR_TEXT(text) tsa_error_text = text;
#else
#define TSA_ERROR_TEXT(text)
#endif

#endif	/* __TSA_UTIL_H__ */
